@@ -31,6 +31,7 @@
 #include <hash-pjw-bare.h>
 #include "x509_int.h"
 #include <common.h>
+#include <gnutls/x509-ext.h>
 #include "verify-high.h"
 
 struct named_cert_st {
@@ -242,16 +243,21 @@ gnutls_x509_trust_list_add_cas(gnutls_x509_trust_list_t list,
 		hash %= list->size;
 
 		/* avoid duplicates */
-		if (flags & GNUTLS_TL_NO_DUPLICATES) {
+		if (flags & GNUTLS_TL_NO_DUPLICATES || flags & GNUTLS_TL_NO_DUPLICATE_KEY) {
 			for (j=0;j<list->node[hash].trusted_ca_size;j++) {
-				if (_gnutls_check_if_same_cert(list->node[hash].trusted_cas[j], clist[i]) != 0) {
+				if (flags & GNUTLS_TL_NO_DUPLICATES)
+					ret = _gnutls_check_if_same_cert(list->node[hash].trusted_cas[j], clist[i]);
+				else
+					ret = _gnutls_check_if_same_key(list->node[hash].trusted_cas[j], clist[i], 1);
+				if (ret != 0) {
 					exists = 1;
 					break;
 				}
 			}
 
 			if (exists != 0) {
-				gnutls_x509_crt_deinit(clist[i]);
+				gnutls_x509_crt_deinit(list->node[hash].trusted_cas[j]);
+				list->node[hash].trusted_cas[j] = clist[i];
 				continue;
 			}
 		}
@@ -677,9 +683,9 @@ int gnutls_x509_trust_list_get_issuer(gnutls_x509_trust_list_t list,
 				      gnutls_x509_crt_t * issuer,
 				      unsigned int flags)
 {
+#ifdef ENABLE_PKCS11
 	int ret;
 
-#ifdef ENABLE_PKCS11
 	if (list->pkcs11_token) {
 		gnutls_datum_t der = {NULL, 0};
 		/* use the token for verification */
@@ -692,7 +698,7 @@ int gnutls_x509_trust_list_get_issuer(gnutls_x509_trust_list_t list,
 		 * persistent. It will be deallocated when the trust list is.
 		 */
 		ret = gnutls_x509_trust_list_add_trust_mem(list, &der, NULL,
-			GNUTLS_X509_FMT_DER, GNUTLS_TL_NO_DUPLICATES, 0);
+			GNUTLS_X509_FMT_DER, GNUTLS_TL_NO_DUPLICATE_KEY, 0);
 		if (ret < 0)
 			return gnutls_assert_val(ret);
 
@@ -753,15 +759,74 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 				  unsigned int flags,
 				  unsigned int *voutput,
 				  gnutls_verify_output_function func)
+{
+	return gnutls_x509_trust_list_verify_crt2(list, cert_list, cert_list_size,
+						  NULL, 0, flags, voutput, func);
+}
+
+/**
+ * gnutls_x509_trust_list_verify_crt2:
+ * @list: The structure of the list
+ * @cert_list: is the certificate list to be verified
+ * @cert_list_size: is the certificate list size
+ * @data: an array of typed data
+ * @elements: the number of data elements
+ * @flags: Flags that may be used to change the verification algorithm. Use OR of the gnutls_certificate_verify_flags enumerations.
+ * @voutput: will hold the certificate verification output.
+ * @func: If non-null will be called on each chain element verification with the output.
+ *
+ * This function will try to verify the given certificate and return
+ * its status. The @verify parameter will hold an OR'ed sequence of
+ * %gnutls_certificate_status_t flags.
+ *
+ * Additionally a certificate verification profile can be specified
+ * from the ones in %gnutls_certificate_verification_profiles_t by
+ * ORing the result of GNUTLS_PROFILE_TO_VFLAGS() to the verification
+ * flags.
+ *
+ * The acceptable @data types are %GNUTLS_DT_DNS_HOSTNAME and %GNUTLS_DT_KEY_PURPOSE_OID.
+ * The former accepts as data a null-terminated hostname, and the latter a null-terminated
+ * object identifier (e.g., %GNUTLS_KP_TLS_WWW_SERVER).
+ * If a DNS hostname is provided then this function will compare
+ * the hostname in the certificate against the given. If names do not match the 
+ * %GNUTLS_CERT_UNEXPECTED_OWNER status flag will be set.
+ * If a key purpose OID is provided and the end-certificate contains the extended key
+ * usage PKIX extension, it will be required to be have the provided key purpose 
+ * or be marked for any purpose, otherwise verification will fail with %GNUTLS_CERT_SIGNER_CONSTRAINTS_FAILURE status.
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ *   negative error value. Note that verification failure will not result to an
+ *   error code, only @voutput will be updated.
+ *
+ * Since: 3.3.8
+ **/
+int
+gnutls_x509_trust_list_verify_crt2(gnutls_x509_trust_list_t list,
+				  gnutls_x509_crt_t * cert_list,
+				  unsigned int cert_list_size,
+				  gnutls_typed_vdata_st *data,
+				  unsigned int elements,
+				  unsigned int flags,
+				  unsigned int *voutput,
+				  gnutls_verify_output_function func)
 {
 	int ret;
 	unsigned int i;
 	uint32_t hash;
 	gnutls_x509_crt_t sorted[DEFAULT_MAX_VERIFY_DEPTH];
+	const char *hostname = NULL, *purpose = NULL;
 
 	if (cert_list == NULL || cert_list_size < 1)
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
+	for (i=0;i<elements;i++) {
+		if (data[i].type == GNUTLS_DT_DNS_HOSTNAME) {
+			hostname = (void*)data[i].data;
+		} else if (data[i].type == GNUTLS_DT_KEY_PURPOSE_OID) {
+			purpose = (void*)data[i].data;
+		}
+	}
+
 	if (!(flags & GNUTLS_VERIFY_DO_NOT_ALLOW_UNSORTED_CHAIN))
 		cert_list = sort_clist(sorted, cert_list, &cert_list_size);
 
@@ -779,6 +844,7 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 	ret = check_if_in_blacklist(cert_list, cert_list_size,
 		list->blacklisted, list->blacklisted_size);
 	if (ret != 0) {
+		*voutput = 0;
 		*voutput |= GNUTLS_CERT_REVOKED;
 		*voutput |= GNUTLS_CERT_INVALID;
 		return 0;
@@ -790,6 +856,7 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 
 		*voutput = _gnutls_pkcs11_verify_crt_status(list->pkcs11_token,
 								cert_list, cert_list_size,
+								purpose!=NULL?purpose:GNUTLS_KP_TLS_WWW_SERVER,
 								flags, func);
 	} else
 #endif
@@ -825,6 +892,24 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 		}
 	}
 
+	/* End-certificate, key purpose and hostname checks. */
+	if (purpose) {
+		ret = _gnutls_check_key_purpose(cert_list[0], purpose);
+		if (ret != 1) {
+			gnutls_assert();
+			*voutput |= GNUTLS_CERT_SIGNER_CONSTRAINTS_FAILURE|GNUTLS_CERT_INVALID;
+		}
+	}
+
+	if (hostname) {
+		ret =
+		    gnutls_x509_crt_check_hostname2(cert_list[0], hostname, flags);
+		if (ret == 0)
+			*voutput |= GNUTLS_CERT_UNEXPECTED_OWNER|GNUTLS_CERT_INVALID;
+	}
+
+	/* CRL checks follow */
+
 	if (*voutput != 0 || (flags & GNUTLS_VERIFY_DISABLE_CRL_CHECKS))
 		return 0;
 
@@ -904,6 +989,7 @@ gnutls_x509_trust_list_verify_named_crt(gnutls_x509_trust_list_t list,
 	unsigned int i;
 	uint32_t hash;
 
+
 	hash =
 	    hash_pjw_bare(cert->raw_issuer_dn.data,
 			  cert->raw_issuer_dn.size);
@@ -912,6 +998,7 @@ gnutls_x509_trust_list_verify_named_crt(gnutls_x509_trust_list_t list,
 	ret = check_if_in_blacklist(&cert, 1,
 		list->blacklisted, list->blacklisted_size);
 	if (ret != 0) {
+		*voutput = 0;
 		*voutput |= GNUTLS_CERT_REVOKED;
 		*voutput |= GNUTLS_CERT_INVALID;
 		return 0;
